//
// やすらかシステム ノード側
//
//	バージョン
//		0.1.2
//
//	使い方：
//		sudo ./yasuraka_node [通知周期(s)] [IPアドレス] [ポート番号]
//

//-- インクルード --
#include <assert.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <netinet/in.h>
#include <pthread.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>
#include <errno.h>

#include <wiringPi.h>

//-- マクロ --
#define	NOTIFICATION_CYCLE	(300)				// 引数で指定が無い場合のサーバへの通知周期(sec)
#define	SERVER_ADDRESS		"192.168.11.11"		// 引数で指定が無い場合のサーバのIPアドレス
#define	SERVER_PORT			(10000)				// 引数で指定が無い場合のサーバのポート番号
#define	GPIO_LED			(18)				// LEDのポート番号
#define	GPIO_SEN			(22)				// 赤外線センサのポート番号


//-- 型定義 --
struct thdata {
	char				letter;
};

//-- プロトタイプ宣言 --
int		main( int, char*[]);
void	*keep_connection(void *thdata);
void	timer_handler( int, siginfo_t*, void*);
void	signal_handler(int sig);

//-- グローバル変数 --
int			sock;
int			exit_request		= 0;					// 終了要求フラグ
int			server_connection	= 0;					// サーバとの接続状態(0=切断,1=接続中)
pthread_t	th_connection;
int			notification_cycle	= NOTIFICATION_CYCLE;	// サーバへの通知周期(sec)
char		server_address[16]	= SERVER_ADDRESS;		// サーバのIPアドレス
int			server_port			= SERVER_PORT;			// サーバのポート番号

//-- 外部参照 --
extern float	get_temperature(void);
extern float	get_humidity(void);


//
// main関数
//
int  main( int argc, char *argv[])	{
	struct sigaction 	action;
	struct sigevent		evp;
	struct itimerspec	ispec;
	timer_t				timerid = 0;
	
	// 引数の処理
	if( argc > 1 && ( !strcmp(argv[1],"--help") || !strcmp(argv[1],"-h") | !strcmp(argv[1],"/?") ) )  {		// 使い方の表示が要求された？
		printf("\nやすらかシステム(ノード) プロトタイプ\n\n");
		printf("使い方：yasuraka_node [通知周期(s)] [IPアドレス] [ポート番号]\n\n");
		exit(EXIT_SUCCESS);
	}
	if( argc > 1 )	{	// サーバへの通知周期が指定された？
		notification_cycle = atoi(argv[1]);
	}
	if( argc > 2 )	{	// サーバのIPアドレスが指定された？
		if( strlen(argv[2]) < 16) {
			printf("length=%d\n", strlen(argv[2]));
			strcpy( server_address, argv[2]);
		}
	}
	if( argc > 3 )	{	// サーバのポート番号が指定された？
		server_port = atoi(argv[3]);
	}
	
	// wiringPiを初期化
	if(wiringPiSetupGpio() == -1)  {
		exit(EXIT_FAILURE);
	}
	pinMode( GPIO_SEN, INPUT);
	pinMode( GPIO_LED, OUTPUT);
	digitalWrite( GPIO_LED, 1);		// 通信異常LEDを点灯
	
	// シグナルハンドラを設定
	if( signal( SIGINT, signal_handler) == SIG_ERR )  {
		exit(EXIT_FAILURE);
	}
	if( signal( SIGTERM, signal_handler) == SIG_ERR )  {
		exit(EXIT_FAILURE);
	}
	if( signal( SIGHUP, signal_handler) == SIG_ERR )  {
		exit(EXIT_FAILURE);
	}
	if( signal( SIGPIPE, signal_handler) == SIG_ERR )  {
		exit(EXIT_FAILURE);
	}
	
	// タイマハンドラを設定
	memset( &action, 0, sizeof(action));
	memset( &evp,	 0, sizeof(evp));
	action.sa_sigaction = timer_handler;
	action.sa_flags 	= SA_SIGINFO | SA_RESTART;
	sigemptyset( &action.sa_mask);
	if( sigaction(SIGRTMIN+1, &action, NULL) < 0 )	{
		perror("sigactionエラー");
		exit(EXIT_FAILURE);
	}
	
	// タイマを設定
	evp.sigev_notify = SIGEV_SIGNAL;
	evp.sigev_signo  = SIGRTMIN + 1;
	if( timer_create( CLOCK_REALTIME, &evp, &timerid) < 0 )  {
		perror("timer_createエラー");
		exit(EXIT_FAILURE);
	}
	ispec.it_interval.tv_sec = 1;
	ispec.it_interval.tv_nsec = 0;
	ispec.it_value.tv_sec = 1;
	ispec.it_value.tv_nsec = 0;
	if(timer_settime(timerid, 0, &ispec, NULL) < 0){
		perror("timer_settimeエラー");
		exit(EXIT_FAILURE);
	}
	
	// 接続維持スレッドを起動
	if( pthread_create( &th_connection, NULL, keep_connection, (void *)NULL) != 0 )  {
		perror("pthread_createエラー");
		exit(EXIT_FAILURE);
	}
	
	/* 接続維持スレッドが終了するのを待つ */
	printf("やすらかシステム(ノード) プロトタイプ\nctrl-Cで終了します。\n");
	printf("サーバへの通知周期=%d秒 サーバのIPアドレス=%s ポート番号=%d\n",
									notification_cycle,
									server_address,
									server_port);
	pthread_join( th_connection, NULL);
	printf("終了します。\n");
	
	exit(EXIT_SUCCESS);
}

//
// 接続維持スレッド
//		サーバとの接続を常に維持する。
void  *keep_connection(void *thdata)  {
	struct thdata		*priv = (struct thdata *)thdata;
	struct sockaddr_in	server;
	
	// サーバとの接続を維持
	while( !exit_request )	{
		if( !server_connection )  {
			// 終了要求フラグが立っていたら終了
			if( close(sock) )  {
				perror("closeエラー");
				errno = 0;
			}
			// ソケットを作成
			sock = socket( AF_INET, SOCK_STREAM, 0);
			if( sock == -1 )  {
				perror("socketエラー");
				errno = 0;
				return;
			}
			// サーバへ接続
			server.sin_family = AF_INET;
			server.sin_port   = htons(server_port);
			server.sin_addr.s_addr = inet_addr(server_address);
			if( connect( sock, (struct sockaddr *)&server, sizeof(server)) == (-1) )  {
				perror("connectエラー");
				errno = 0;
			} else {
				digitalWrite(GPIO_LED, 0);		// 通信異常LEDを消灯
				server_connection = 1;
				printf("サーバへ接続しました。\n");
			}
		}
		sleep(1);
	}
	return;
}

//
// タイマーハンドラ
//		1秒毎にコールされる。
void  timer_handler( int signum, siginfo_t *info, void *ctxt)  {
	static unsigned long	sec_cnt = 0;		// 経過時間カウンタ(sec)
	static unsigned long	moving_cnt = 0;		// 動き時間継続カウンタ(sec)
	char					buf[64];
	ssize_t					res;
	float					temperature;
	float					humidity;
	
	sec_cnt ++;		// 経過時間をインクリメント
	
	// 動きがあれば動き時間継続カウンタをインクリメント
	if( digitalRead(GPIO_SEN) && moving_cnt < LONG_MAX )  {
		moving_cnt++;
	}
	
	// 通知周期毎の処理
	if( !(sec_cnt % notification_cycle) )  {		// 通知のタイミングか？
		temperature	= get_temperature();
		humidity	= get_humidity();
		// サーバへ通知
		printf("動き継続時間=%lu(s) 温度=%f(℃) 湿度=%f(%)\n", moving_cnt, temperature, humidity);
		if( server_connection )  {
			sprintf(buf, "moving=%lu;temperature=%f;humidity=%f\n", moving_cnt, temperature, humidity);
			res = write( sock, buf, strlen(buf));
			printf("送信結果=%d, エラーコード=%d\n", res, errno);
			if( res == (-1) )  {	// wirte失敗？
				digitalWrite(GPIO_LED, 1);		// 通信異常LEDを点灯
				server_connection = 0;
				errno = 0;
			}
		} else {
			printf("サーバと接続していません。\n");
		}
		moving_cnt = 0;
	}
	return;
}

//
// シグナルハンドラ
//		全シグナルをハンドリングする。
void  signal_handler(int sig)  {
	// ctrl-Cが押されたら接続維持スレッドを終了
	printf("signal_handler=%d\n", sig);
	if( sig == SIGINT )  {
		exit_request = 1;
	}
	if( sig == SIGTERM )  {
		exit(EXIT_SUCCESS);
	}
}
